# Higgs example

In this example we will train a Logistic Regression model on the HIGGS dataset, using both sparkML and snap-ml-spark. 

The HIGGS dataset is avaliable in the [UCI machine learning repository](https://archive.ics.uci.edu/ml/datasets/HIGGS). 

The problem is described as follows:

> This is a classification problem to distinguish between a signal process which produces Higgs bosons and a background process which does not.

The corresponding citation for the dataset is the following paper:

> Baldi, P., P. Sadowski, and D. Whiteson. “Searching for Exotic Particles in High-energy Physics with Deep Learning.” Nature Communications 5 (July 2, 2014).

## Download the Data

Download and decompress the data from the LIBSVM repository:
```bash
mkdir data
cd data
wget https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/HIGGS.bz2
bunzip2 HIGGS.bz2
cd ../
```

Finally, one should see a file named HIGGS in the `data/` directory

## Data Pre-processing
The data is in libsvm format which is not really beneficial since this dataset is 95% dense. Therefore, we perform 
the following pre-processing steps.

Firstly, we split out data to train/test sets.

Next, we convert them to dense representation since as we have mentioned the dataset is 95% dense

Then, also normalize our data. Please note that it is always recommended to normalize your data when using the `snap_ml_spark` library in order to achieve the maximum performance. 

We write our preprocessed train and test data to libsvm format.

And finally, we also write the train set to dense snap format which will be used only for the `snap_ml_spark` Logistic Regression training in order to achieve higher performance.

The produced preprocessed files should be the following:
```bash
ls -l data/
HIGGS 
HIGGS-test.libsvm  
HIGGS-train_dense.snap  
HIGGS-train.libsvm
```

The entire pre-processing code can be run by executing `python preprocess-higgs.py`


## Training and Evaluating a Logistic Regression Model

After running the data pre-processing steps above, we are ready to perform our benchmarking:

Firstly we need to import the pySpark modules, set the configuration and disable manually the LOG INFO printed by spark.
The configuration can vary depending on your Spark cluster.

Then, our next step is to define the data paths.
Please note that if you wish to run Spark on a real distributed environment, then you must set the absolute data paths of the train/test sets and
also have access to the data from all the connected machines that will be used (by replicating the data or by providing a distributed file system).

Next, starting from the `snap_ml_spark` library first, our first step is to read the snap_dense train and libsvm test data.
As we mentioned earlier, for `snap_ml_spark` we read the data into snap format in order to achieve the highiest possible performance.

After the initialization, all we need to do is to call the `fit()` function in order to train our dual Logistic Regression model (we also time it).

As a last step, we make our predictions on the test data, we compute the accuracy and the logistic loss metrics.

We perform the above steps by enabling or disabling GPUs.

Finally, we perform the same steps using the `Spark ML` library and the same pySpark configuration (We use the default values of the Spark ML constructor).

## How to run examples

One can run the entire example by executing:
```bash
spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-higgs/example-higgs.py --data_path path/to/data

Note: By default the location of --data_path is current directory. Parent directory of the directory 'data/' is to be given as parameter to --data_path option.

Command to run example using GPUs:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-higgs/example-higgs.py --use_gpu

Command to run example with specified number of partitions:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-higgs/example-higgs.py --num_partitions 8

Note: Default value of num_partitions is 4. num_partitions can be 8, if user wants to use 8 GPUs.
```
One should see the following output on a **P9+V100 machine with 4 gpus**:
```
[...]
snapML accuracy: 0.6431 , GPU time: 15.79 s
sparkML lib accuracy: 0.6431 ,    time: 142.02 s
```
Please keep in mind that the `snap_ml_spark` CPU algorithm converges significantly faster than the snap_ml_spark GPU algorithm with respect to the number of epochs.
The reported times can vary depending on system architecture.

## Duality

In this example we provide an additional example where we train our Logistic Regression model on the same dataset but using the Primal and the Dual formulations.
The steps are identical with the aforementioned steps.
One should run the following python script 

```bash
python preprocess-higgs-duality.py
```

to preprocess the data and the following command to run the duality example
```bash
spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-higgs/example-higgs-duality.py --data_path path/to/data

Note: By default the location of --data_path is current directory. Parent directory of the directory 'data/' is to be given as parameter to --data_path option.

Command to run example using GPUs:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-higgs/example-higgs-duality.py --use_gpu

Command to run example with specified number of partitions:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-higgs/example-higgs-duality.py --num_partitions 8

Note: Default value of num_partitions is 4. num_partitions can be 8, if user wants to use 8 GPUs.
```
&copy; Copyright IBM Corporation 2018, 2019
