# Webspam example

In this example we will train a Logistic Regression model on the Webspam dataset, using both `sparkML` and `snap-ml-spark`. 

The Webspam dataset is avaliable in the [LIBSVM repository](https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary.html). 

## Download the Data

Download and decompress the data from the LIBSVM repository:
```bash
mkdir data
cd data
wget https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/webspam_wc_normalized_trigram.svm.bz2
bunzip2 webspam_wc_normalized_trigram.svm.bz2
cd ../
```

Finally, one should see a file named `webspam_wc_normalized_trigram.svm` in the `data/` directory

## Data Pre-processing
We perform the following pre-processing steps.


Firstly, we split out data to train/test sets
```python
# Load the data
from sklearn.datasets import load_svmlight_file
filename = "data/webspam_wc_normalized_trigram.svm"
X,y = load_svmlight_file(filename)

y[y == -1] = 0

# Make the train-test split
from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25, random_state=42)
```
Note that we also convert the labels from -1 to 0 in order to be able to train the dataset using the sparkML library (It does not support -1 labels).

Next, we normalize (by example) our data. Please note that it is always recommended to normalize your data when using the `snap_ml_spark` library in order to achieve the maximum performance. 
```python
from sklearn.preprocessing import normalize
X_train = normalize(X_train, axis=1, norm='l1')
X_test  = normalize(X_test,  axis=1, norm='l1')
```

We write our preprocessed train and test data to libsvm format
```python
# Save to libsvm files (index must start from 1)
from sklearn.datasets import dump_svmlight_file
dump_svmlight_file(X_train,y_train, filename+'-train.libsvm', zero_based=False)
dump_svmlight_file(X_test,y_test, filename+'-test.libsvm', zero_based=False)
```

And finally, we also write the train set to snap transposed format in order to train on the snap-ml-spark Primal Logistic Regression.
```python
from snap_ml_spark.Utils import dump_to_snap_format
dump_to_snap_format(X_train,y_train, filename+'-train_sparse.snap.t', transpose=True)
```


The produced preprocessed files should be the following:
```bash
ls -l data/
webspam_wc_normalized_trigram.svm
webspam_wc_normalized_trigram.svm-test.libsvm
webspam_wc_normalized_trigram.svm-train.libsvm
webspam_wc_normalized_trigram.svm-train_sparse.snap.t
```

The entire pre-processing code can be run by executing `python preprocess-webspam.py`

## Training and Evaluating a Logistic Regression Model

After running the data pre-processing steps above, we are ready to perform our benchmarking:

Firstly we need to import the pySpark modules, set the configuration and disable manually the LOG INFO printed by spark
```python
from pyspark import SparkConf, SparkContext
from pyspark.sql.session import SparkSession

conf = SparkConf()\
.set("spark.solver.count",args.num_partitions)\
.set("spark.cores.max",args.num_partitions)\
.setAppName("Spark-criteoKaggle2014-SparkComparison")\
.set("spark.executor.extraJavaOptions","-XX:MaxDirectMemorySize=8192m")\
.set("spark.driver.maxResultSize",8192)
# Create spark config
sc = SparkContext(conf=conf)
# Configure Spark
spark = SparkSession(sc)

# disable spark INFO logging - comment out to enable
log4j = sc._jvm.org.apache.log4j
log4j.LogManager.getRootLogger().setLevel(log4j.Level.ERROR)
```
The configuration can vary depending on your Spark cluster

Then, our next step is to define the data paths.
Please note that if you wish to run Spark on a real distributed environment, then you must set the absolute data paths of the train/test sets and
also have access to the data from all the connected machines that will be used (by replicating the data or by providing a distributed file system).
```python
filename = 'data/webspam_wc_normalized_trigram.svm'
train_filename_sparse_snap_t = filename + '-train_sparse.snap.t'
train_filename = filename + '-train.libsvm'
test_filename = filename + '-test.libsvm'
```

Next, starting from the `snap_ml_spark` library first, our first step is to read the snap_dense (we need to read the transpose format to train on the Primal objective function) train and libsvm test data:
```python
from snap_ml_spark import DatasetReader
from snap_ml_spark import LogisticRegression as snapml_LogisticRegression
from snap_ml_spark.Metrics import accuracy, logisticLoss
n_features_ = 16609143
train_data_t = DatasetReader().setFormat("snap").load(train_filename_sparse_snap_t)
test_data = DatasetReader().setFormat("libsvm").setNumFt(n_features_).load(test_filename)
```

Then, we initialize our `snap_ml_spark` LogisticRegression constructor by setting the number of coordinate descent epochs to 19 and the snapml regularizer to 1.0.
We leave the rest to the default values (GPUs disabled, number of threads=1). After the initialization, all we need to do is to call the `fit()` function in order to train our dual Logistic Regression model (we also time it):
```python
snapml_regularizer = 1.0
snapml_lr = snapml_LogisticRegression(max_iter=19, regularizer=snapml_regularizer, verbose=False, dual=False, use_gpu=False)

snapml_t0 = time.time()
snapml_lr.fit(train_data_t)
snapml_time_cpu_primal = time.time() - snapml_t0
```
Please note that we set the `dual` to `False` since here we are optimizing the Primal Logistic Regression objective function which makes sense considering the dimensions of the webspam dataset (350K X 16M).

As a last step, we make our predictions on the test data, we compute the accuracy and the logistic loss metrics and we print off the results (with the time needed to achieve them):
```python
pred = snapml_lr.predict(test_data)

snapml_accuracy_cpu_primal = accuracy(pred)

print('snapML CPU accuracy: %.4f' %snapml_accuracy_cpu_primal, ", CPU time: %.2f" % snapml_time_cpu_primal, 's')
```

Then, we perform the same steps for the GPU snap_ml_spark Primal Logistic Regression algorithm (GPUs are enabled this time). Note that the `use_gpu` parameter in the Logistic Regression constructor is set to `False` and thus we need to set it to `True`.
```python
snapml_lr = snapml_LogisticRegression(max_iter=70, regularizer=snapml_regularizer, verbose=False, dual=False, use_gpu=True) # iter 70

snapml_t0 = time.time()
snapml_lr.fit(train_data_t)
snapml_time_gpu_primal = time.time() - snapml_t0

pred = snapml_lr.predict(test_data)

snapml_accuracy_gpu_primal = accuracy(pred)

print('snapML GPU primal accuracy: %.4f' %snapml_accuracy_gpu_primal, ", GPU time: %.2f" % snapml_time_gpu_primal, 's')
```

Finally, we perform the same steps using the `Spark ML` library and the same pySpark configuration (We use the default values of the Spark ML constructor but we disable the feature scaling and the standardization in order to make the algorithm equivalent to snap-ml-spark)
```python
from pyspark.ml.classification import LogisticRegression as sparkml_LogisticRegression

# Load training data
train_data = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(train_filename)
test_data = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(test_filename)

n_examples = train_data.count()

sparkml_lr = sparkml_LogisticRegression(fitIntercept=False, regParam=snapml_regularizer/n_examples, standardization=False)

sparkml_t0 = time.time()
sparkml_lr_model = sparkml_lr.fit(train_data)
sparkml_time = time.time() - sparkml_t0

predictions = sparkml_lr_model.transform(test_data)

from pyspark.ml.evaluation import MulticlassClassificationEvaluator
evaluator = MulticlassClassificationEvaluator(labelCol="label", predictionCol="prediction", metricName="accuracy")
sparkml_accuracy = evaluator.evaluate(predictions)

print('Spark ml', evaluator.getMetricName(),'=', sparkml_accuracy,", time: %.2f" % sparkml_time, 's')
```

## Results

After the final training and prediction on the train/test data using the Spark ML, we finally print off all the results:
```python
print('\nsnapML  CPU primal accuracy: %.4f' %snapml_accuracy_cpu_primal, ", CPU time: %.2f" % snapml_time_cpu_primal, 's')
print('snapML  GPU primal accuracy: %.4f' %snapml_accuracy_gpu_primal, ", GPU time: %.2f" % snapml_time_gpu_primal, 's')
print('sparkML lib accuracy: %.4f' % sparkml_accuracy, ",    time: %.2f" % sparkml_time, 's')
```

One can run the entire example by executing:
```bash
spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-webspam/example-webspam.py --data_path path/to/data

Note: By default the location of --data_path is current directory. Parent directory of the directory 'data/' is to be given as parameter to --data_path option.

Command to run example using GPUs:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-webspam/example-webspam.py --use_gpu

ommand to run example with specified number of partitions:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-webspam/example-webspam.py --num_partitions

Note: Default value of num_partitions is 4. num_partitions can be 8, if user wants to use 8 GPUs.
```
One should see the following output on a **P9+V100 machine with 4 gpus**:
```
[...]
snapML  GPU primal accuracy: 0.9931 , GPU time: 46.75 s
sparkML lib accuracy: 0.9773 ,            time: 5166.13 s
```
Please keep in mind that the `snap_ml_spark` CPU algorithm converges significantly faster than the snap_ml_spark GPU algorithm with respect to the number of epochs.
The reported times can vary depending on system architecture.

&copy; Copyright IBM Corporation 2018, 2019
